@@ -1,4 +1,6 @@
 # -*- coding: utf-8 -*-
+import logging
+from functools import partial
 from App.config import getConfiguration
 
 from zope.component import queryMultiAdapter
@@ -9,33 +11,62 @@ from collective.geo.kml.interfaces import IFeature
 
 from fastkml import kml, styles
 
-from shapely.geometry import asShape
+try:
+    from shapely.geometry import asShape
+except:
+    from pygeoif import as_shape as asShape
+
 
 from collective.geo.fastkml import MessageFactory as _
-from zope.i18n import translate
+
+logger = logging.getLogger('collective.geo.fastkml')
 
 
 class FastKMLBaseDocument(KMLBaseDocument):
 
     def anchorsnippet(self, link):
-        snippettext = self.context.translate(_(u'See the original resource'))
-        return '''<p class="placemark-url">
-            <a href="%s">%s</a></p>''' % (link, snippettext)
+        return (
+            '<p class="placemark-url">'
+                '<a href="%s">%s</a>'
+            '</p>'
+        ) % (link, self.context.translate(_(u'See the original resource')))
 
     def get_kml(self):
 
-        k = kml.KML()
+        # Some programs consuming kml cannot handle kml namespaces. For those
+        # programs the supress-kml-namespace parameter may be used.
+        # e.g. .../@@kml-document?suppress-kml-prefix
+        if self.request.get('supress-kml-prefix', None) is not None:
+            namespace = ''  # no kml namespace prefixes
+        else:
+            namespace = None  # use default
+
+        KML = partial(kml.KML, ns=namespace)
+        Document = partial(kml.Document, ns=namespace)
+        Placemark = partial(kml.Placemark, ns=namespace)
+        Style = partial(styles.Style, ns=namespace)
+        IconStyle = partial(styles.IconStyle, ns=namespace)
+        LineStyle = partial(styles.LineStyle, ns=namespace)
+        PolyStyle = partial(styles.PolyStyle, ns=namespace)
+        UntypedExtendedData = partial(kml.UntypedExtendedData, ns=namespace)
+        UntypedExtendedDataElement = partial(
+            kml.UntypedExtendedDataElement, ns=namespace
+        )
+
+        k = KML()
+
         ## make sure description field is encoded properly
-        desc = unicode(self.description, \
-                       'utf-8').encode('ascii', 'xmlcharrefreplace')
-        doc = kml.Document(name=self.name, description=desc)
+        doc = Document(
+            name=unicode(self.name, 'utf-8'),
+            description=unicode(self.description, 'utf-8')
+        )
         k.append(doc)
-        docstyle = styles.Style(id="defaultStyle")
-        istyle = styles.IconStyle(scale=self.marker_image_size,
+        docstyle = Style(id="defaultStyle")
+        istyle = IconStyle(scale=self.marker_image_size,
                 icon_href=self.marker_image,
                 color=self.polygoncolor)
-        lstyle = styles.LineStyle(color=self.linecolor, width=self.linewidth)
-        pstyle = styles.PolyStyle(color=self.polygoncolor)
+        lstyle = LineStyle(color=self.linecolor, width=self.linewidth)
+        pstyle = PolyStyle(color=self.polygoncolor)
         docstyle.append_style(istyle)
         docstyle.append_style(lstyle)
         docstyle.append_style(pstyle)
@@ -45,11 +76,11 @@ class FastKMLBaseDocument(KMLBaseDocument):
             if feature.lead_image():
                 description += feature.lead_image()
             if feature.description:
-                description += unicode(feature.description, \
-                                  'utf-8').encode('ascii', 'xmlcharrefreplace')
+                description += unicode(feature.description, 'utf-8')
             if feature.item_url:
                 description += self.anchorsnippet(feature.item_url)
-            pm = kml.Placemark(name=feature.name, description=description)
+            name = unicode(feature.name, 'utf-8')
+            pm = Placemark(name=name, description=description)
             shape = {'type': feature.geom.type,
                 'coordinates': feature.geom.coordinates}
             try:
@@ -57,30 +88,50 @@ class FastKMLBaseDocument(KMLBaseDocument):
             except:
                 continue
             pm.author = feature.author['name']
+            if feature.item_url:
+                pm.link = unicode(feature.item_url, 'utf-8')
             if feature.use_custom_styles:
-                pms = styles.Style()
+                pms = Style()
                 if feature.geom.type in ['Point', 'MultiPoint']:
-                    istyle = styles.IconStyle(scale=feature.marker_image_size,
+                    istyle = IconStyle(scale=feature.marker_image_size,
                         icon_href=feature.marker_image,
                         color=feature.polygoncolor)
                     pms.append_style(istyle)
                 elif feature.geom.type in ['LineString', 'MultiLineString',
                                             'Polygon', 'MultiPolygon']:
-                    lstyle = styles.LineStyle(color=feature.linecolor,
+                    lstyle = LineStyle(color=feature.linecolor,
                                     width=feature.linewidth)
                     pms.append_style(lstyle)
                 if feature.geom.type in ['Polygon', 'MultiPolygon']:
-                    pstyle = styles.PolyStyle(color=feature.polygoncolor)
+                    pstyle = PolyStyle(color=feature.polygoncolor)
                     pms.append_style(pstyle)
                 pm.append_style(pms)
             else:
                 pm.styleUrl = "#defaultStyle"
+            try:
+                extended_data = UntypedExtendedData()
+                for element in feature.extended_data:
+                    extended_data.elements.append(
+                        UntypedExtendedDataElement(
+                            name=element.name,
+                            value=element.value,
+                            display_name=element.display_name
+                        )
+                    )
+                if extended_data.elements:
+                    pm.extended_data = extended_data
+            except AttributeError as e:
+                logger.debug(e.message)
             doc.append(pm)
-        if getConfiguration().debug_mode:
-            xml = '<?xml version="1.0" encoding="UTF-8"?>' + \
-                    k.to_string(prettyprint=True)
-        else:
-            xml = '<?xml version="1.0" encoding="UTF-8"?>' + k.to_string()
+
+        pretty_print =  any((
+            getConfiguration().debug_mode,
+            self.request.get('pretty-print', None) is not None
+        ))
+        
+        xml = u'<?xml version="1.0" encoding="UTF-8"?>' + \
+                k.to_string(prettyprint=pretty_print)
+        
         return xml
 
     def __call__(self):
@@ -91,10 +142,6 @@ class FastKMLBaseDocument(KMLBaseDocument):
         self.request.response.setHeader('Content-Type',
             'application/vnd.google-earth.kml+xml; charset=utf-8')
         xml = self.get_kml()
-        try:
-            xml = xml.decode('utf-8', 'ignore')
-        except:
-            pass
         return xml.encode('utf-8')
 
 
@@ -112,6 +159,18 @@ class FastBrainPlacemark(BrainPlacemark):
         else:
             return ''
 
+    @property
+    def item_type(self):
+        return self.context.portal_type
+
+    @property
+    def item_url(self):
+        return self.context.getURL()
+
+    @property
+    def extended_data(self):
+        return []
+
 
 class KMLDocument(FastKMLBaseDocument):
 
@@ -127,21 +186,17 @@ class KMLFolderDocument(FastKMLBaseDocument):
 
     @property
     def features(self):
-        for brain in self.context.getFolderContents():
-            try:
-                if brain.zgeo_geometry['coordinates']:
-                    yield FastBrainPlacemark(brain, self.request, self)
-            except:
+        for item in self.context.values():
+            feature = queryMultiAdapter((item, self.request), IFeature)
+            if not feature:
                 continue
+            yield feature
+
 
 class KMLTopicDocument(FastKMLBaseDocument):
 
     @property
     def features(self):
         for brain in self.context.queryCatalog():
-            try:
-                if brain.zgeo_geometry['coordinates']:
-                    yield FastBrainPlacemark(brain, self.request, self)
-            except:
-                continue
+            yield FastBrainPlacemark(brain, self.request, self)
 
